# Copyright (C) 2020, 2021, 2022, 2023 Serghei Iakovlev <egrep@protonmail.ch>
#
# This file is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This file is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this file.  If not, see <https://www.gnu.org/licenses/>.

import sys
from argparse import ArgumentParser, Namespace
from unittest import mock

from gstore import args, env


def test_none_args(monkeypatch):
    """Show help message and return None if gstore was called without any
    argument and environment variables were not enough to start gstore.
    """
    def mock_lookup_token():
        return None

    monkeypatch.setattr('sys.argv', ['gstore'])
    monkeypatch.setattr(
        env,
        'lookup_token',
        mock_lookup_token
    )

    with mock.patch.object(ArgumentParser, 'print_help') as mock_help:
        assert args.argparse() is None
        mock_help.assert_called_once_with(sys.stderr)


def test_only_token(monkeypatch):
    """Token is enough to run gstore without arguments."""
    def mock_lookup_token():
        return 'secret'

    monkeypatch.setattr('sys.argv', ['gstore'])
    monkeypatch.setattr(
        env,
        'lookup_token',
        mock_lookup_token
    )

    assert isinstance(args.argparse(), Namespace)



# Copyright (C) 2020, 2021, 2022, 2023 Serghei Iakovlev <egrep@protonmail.ch>
#
# This file is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This file is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this file.  If not, see <https://www.gnu.org/licenses/>.

#
# -- Utils -----------------------------------------------------
#

import codecs
import os
import re


def read_file(filepath):
    """Read content from a UTF-8 encoded text file."""
    with codecs.open(filepath, 'rb', 'utf-8') as file_handle:
        return file_handle.read()


def find_version(meta_file):
    """Extract ``__version__`` from meta_file."""
    here = os.path.abspath(os.path.dirname(__file__))
    contents = read_file(os.path.join(here, meta_file))

    meta_match = re.search(
        r"^__version__\s+=\s+['\"]([^'\"]*)['\"]",
        contents,
        re.M
    )

    if meta_match:
        return meta_match.group(1)
    raise RuntimeError(
        'Unable to find __version__ string in package meta file')


#
# -- Project information -----------------------------------------------------
#

# General information about the project.
project = u"gstore"
copyright = u"2020, 2021 Serghei Iakovlev"
author = u"Serghei Iakovlev"

#
# -- General configuration ---------------------------------------------------
#

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.doctest",
    "sphinx.ext.intersphinx",
    "sphinx.ext.todo",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix of source filenames.
source_suffix = ".rst"

# Allow non-local URIs so we can have images in CHANGELOG etc.
suppress_warnings = ["image.nonlocal_uri"]

# The master toctree document.
master_doc = "index"

# The version info
# The short X.Y version.
release = find_version('../gstore/__init__.py')
version = release.rsplit(u".", 1)[0]
# The full version, including alpha/beta/rc tags.

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ["_build"]

# The reST default role (used for this markup: `text`) to use for all
# documents.
default_role = "any"

# If true, '()' will be appended to :func: etc. cross-reference text.
add_function_parentheses = True

#
# -- Options for extlinks ----------------------------------------------------
#
extlinks = {
    "pypi": ("https://pypi.org/project/%s/", ""),
}

#
# -- Options for intersphinx -------------------------------------------------
#
intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "sphinx": ("https://www.sphinx-doc.org/en/master", None),
}

#
# -- Options for TODOs -------------------------------------------------------
#
todo_include_todos = True

# -- Options for HTML output ----------------------------------------------

# html_favicon = None

html_theme = "furo"
html_title = "Gstore"

html_theme_options = {}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# If false, no module index is generated.
html_domain_indices = True

# If false, no index is generated.
html_use_index = True

# If true, the index is split into individual pages for each letter.
html_split_index = False

# If true, links to the reST sources are added to the pages.
html_show_sourcelink = False

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
# html_use_opensearch = ''

# Output file base name for HTML help builder.
htmlhelp_basename = "gstoredoc"

#
# -- Options for manual page output ---------------------------------------
#

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    ("index", project, u"Gstore Documentation", [author], 1)
]

#
# -- Options for Texinfo output -------------------------------------------
#

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        "index",
        project,
        u"Gstore Documentation",
        author,
        project,
        "Synchronize GitHub repositories made easy.",
        "Miscellaneous",
    )
]



# Copyright (C) 2020, 2021, 2022, 2023 Serghei Iakovlev <egrep@protonmail.ch>
#
# This file is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This file is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this file.  If not, see <https://www.gnu.org/licenses/>.

"""Command line argument parsing methods for gstore."""

import os
import sys
import textwrap as _textwrap
from argparse import ArgumentParser, HelpFormatter, Namespace, SUPPRESS

from gstore import __copyright__, __version__
from gstore import env


class LineBreaksFormatter(HelpFormatter):
    """
    ArgParse help formatter that allows line breaks in the usage messages
    and argument help strings.

    Normally to include newlines in the help output of argparse, you have
    use argparse.RawDescriptionHelpFormatter. However, this means raw text is
    enabled everywhere, and not just for specific help entries where we may
    need it.

    This help formatter allows for you to optional enable/toggle raw text on
    individual menu items by prefixing the help string with 'n|'."""
    def _fill_text(self, text, width, indent) -> str:
        text = self._whitespace_matcher.sub(' ', text).strip()
        paragraphs = text.split('|n ')

        multiline_text = ''

        for paragraph in paragraphs:
            formatted_paragraph = _textwrap.fill(
                paragraph,
                width,
                initial_indent=indent,
                subsequent_indent=indent
            ) + '\n'

            multiline_text = multiline_text + formatted_paragraph

        return multiline_text


def get_version_str() -> str:
    """A helper function to format version info."""
    # pylint: disable=consider-using-f-string
    version = '''
    {prog}s {version}|n
    {copy}.|n
    This is free software; see the source for copying conditions.  There is NO|n
    warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
    '''.format(  # noqa E501
        prog='%(prog)',
        version=__version__,
        copy=__copyright__,
    )
    return version


def parser_add_positionals(parser: ArgumentParser) -> ArgumentParser:
    """Add positional parameters group to a parser."""
    pgroup = parser.add_argument_group('Positional parameters')

    pgroup.add_argument('target', nargs='?', type=str,
                        default=env.get_target() or os.getcwd(),
                        help='Base target to sync repos (e.g. folder on disk)')

    return parser


def parser_add_options(parser: ArgumentParser) -> ArgumentParser:
    """Add options group to a parser."""
    token = env.lookup_token()

    ogroup = parser.add_argument_group('Options')

    ogroup.add_argument('-h', '--help', action='help', default=SUPPRESS,
                        help='Print this help message and quit')

    ogroup.add_argument('--token', dest='token', default=token, type=str,
                        help='An authentication token for GitHub API requests')

    ogroup.add_argument('--host', dest='host', default=env.lookup_token(),
                        type=str, help='The GitHub API hostname')

    ogroup.add_argument('-o', '--org', dest='org', action='append', type=str,
                        help='Organization to sync (all if not provided)')

    repo_help = ('Limit sync to the specified repository, ' +
                 'otherwise sync all repositories (format "org:repo")')
    ogroup.add_argument('-r', '--repo', dest='repo', action='append',
                        type=str, help=repo_help)

    ogroup.add_argument('-v', '--verbose', dest='verbose', action='store_true',
                        help='Increase output verbosity')

    jobs_help = ('specifies the number of jobs (sync processes) ' +
                 'to run simultaneously')
    ogroup.add_argument('-j', '--jobs', dest='jobs', default=None, type=int,
                        action='store', nargs='?', help=jobs_help)

    quiet_help = 'Silence any informational messages, but not error ones'
    ogroup.add_argument('-q', '--quiet', dest='quiet', action='store_true',
                        help=quiet_help)

    ogroup.add_argument('-V', '--version', action='version',
                        help="Print program's version information and quit",
                        version=get_version_str())

    dumpversion_help = ("Print the version of the program and don't " +
                        'do anything else')
    ogroup.add_argument('-dumpversion', action='version',
                        help=dumpversion_help, version=__version__)

    return parser


def argparse() -> Namespace or None:
    """
    The function initializes command line arguments parser.

    This function will show help message and return None if gstore was called
    without any argument and environment variables were not enough to start
    gstore.

    :return: The list of parsed arguments or None in case of any error.
    :rtype: :class:`argparse.Namespace` or None
    """
    parser = ArgumentParser(
        description='Synchronize GitHub repositories of your organizations.',
        usage='%(prog)s [options] [[--] target]',
        formatter_class=LineBreaksFormatter,
        add_help=False,
    )

    parser_add_positionals(parser)
    parser_add_options(parser)

    if len(sys.argv) == 1 and env.lookup_token() is None:
        parser.print_help(sys.stderr)
        return None

    return parser.parse_args()



# Copyright (C) 2020, 2021, 2022, 2023 Serghei Iakovlev <egrep@protonmail.ch>
#
# This file is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This file is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this file.  If not, see <https://www.gnu.org/licenses/>.

"""Data classes used by various operations within gstore."""

from dataclasses import dataclass


@dataclass
class Organization:
    """Represents Organization data class."""

    login: str


@dataclass
class Repository:
    """Represents Repository data class."""

    name: str
    org: Organization



# Copyright (C) 2020, 2021, 2022, 2023 Serghei Iakovlev <egrep@protonmail.ch>
#
# This file is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This file is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this file.  If not, see <https://www.gnu.org/licenses/>.

"""Logging initialization for the gstore command-line tool."""

import logging
import sys


class DebugFilter(logging.Filter):
    """Extended standard logging Filter to filer only DEBUG messages."""

    def filter(self, record: logging.LogRecord) -> bool:
        """
        Only messages with record level DEBUG can pass
        for messages with another level an extra handler is used.

        :param tuple record: logging message record
        :returns: True|False
        :rtype: bool
        """
        return record.levelno == logging.DEBUG


class InfoFilter(logging.Filter):
    """Extended standard logging Filter to filer only INFO messages."""

    def filter(self, record: logging.LogRecord) -> bool:
        """
        Only messages with record level INFO can pass
        for messages with another level an extra handler is used.

        :param tuple record: logging message record
        :returns: True|False
        :rtype: bool
        """
        return record.levelno == logging.INFO


def setup_logger(verbose=False, quiet=False) -> logging.Logger:
    """Setup and return the root logger object for the application.

    :param bool verbose: Enable debug logging
    :param bool quiet: Disable info logging
    :rtype: :class:`logging.Logger`
    """

    root = logging.getLogger('gstore')
    root.setLevel(logging.ERROR if quiet else logging.DEBUG)

    formatter = logging.Formatter(
        '%(asctime)s %(name)s[%(process)d]: %(levelname)s: %(message)s',
        '%b %d %X',
    )

    specs = (
        {'on': verbose, 'filter': DebugFilter()},
        {'on': not quiet, 'level': logging.INFO, 'filter': InfoFilter()},
        {'on': True, 'stream': sys.stderr, 'level': logging.ERROR},
    )

    for spec in specs:
        if spec['on']:
            handler = logging.StreamHandler(spec.get('stream', sys.stdout))
            handler.setLevel(spec.get('level', logging.DEBUG))
            handler.setFormatter(formatter)

            if spec.get('filter'):
                handler.addFilter(spec['filter'])

            root.addHandler(handler)

    return root



# Copyright (C) 2020, 2021, 2022, 2023 Serghei Iakovlev <egrep@protonmail.ch>
#
# This file is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This file is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this file.  If not, see <https://www.gnu.org/licenses/>.

import git

from gstore import exceptions


def test_parse_git_errors():
    """Call gstore.exceptions.parse_git_errors() will format error message."""

    actual = ("Your configuration specifies to merge with the ref " +
              "'refs/heads/master' from the remote, but no such ref " +
              "was fetched.")

    expected = ("Your configuration specifies to merge with the ref " +
                "'refs/heads/master' from the remote, but no such ref " +
                "was fetched")

    error = git.CommandError('git', stdout=actual)
    assert exceptions.parse_git_errors(error) == [expected]

    error = git.CommandError('git', stderr=actual)
    assert exceptions.parse_git_errors(error) == [expected]

    error = git.CommandError('git', stdout=actual, stderr=actual)
    assert exceptions.parse_git_errors(error) == [expected, expected]

    error = git.CommandError('git')
    assert exceptions.parse_git_errors(error) == []



# Copyright (C) 2020, 2021, 2022, 2023 Serghei Iakovlev <egrep@protonmail.ch>
#
# This file is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This file is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this file.  If not, see <https://www.gnu.org/licenses/>.

import pytest
from github import Github
from github.GithubException import UnknownObjectException

from gstore.client import Client
from gstore.repo import Organization


class MicroMock:
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)

    def __iter__(self):
        return iter([])


class MockOrganization:
    @staticmethod
    def get_repo(name):
        return MicroMock(name=name)

    @staticmethod
    def get_repos(*args, **kwargs):
        return MicroMock(totalCount=0)


class MockUser:
    login = 'mock_user'

    @staticmethod
    def get_orgs():
        return MicroMock(totalCount=0)


@pytest.fixture
def mock_user(monkeypatch):
    """github.Github.get_user() mocked to return fake user."""
    def mock_get_user(*args, **kwargs):
        return MockUser()

    monkeypatch.setattr(
        Github,
        'get_user',
        mock_get_user
    )


@pytest.fixture
def mock_micro_organization(monkeypatch):
    """github.Github.get_organization() mocked to return fake organization
    object without methods.
    """
    def mock_get_organization(_, name):
        return MicroMock(login=name)

    monkeypatch.setattr(
        Github,
        'get_organization',
        mock_get_organization
    )


@pytest.fixture
def mock_organization(monkeypatch):
    """github.Github.get_organization() mocked to return fake organization."""
    def mock_get_organization(*args, **kwargs):
        return MockOrganization()

    monkeypatch.setattr(
        Github,
        'get_organization',
        mock_get_organization
    )


@pytest.fixture
def mock_orgs_iter(monkeypatch):
    """MicroMock.__iter__() mocked to return fake list of organizations."""
    def mock_iter(*args, **kwargs):
        return iter(
            [MicroMock(login='awesome'), MicroMock(login='company')]
        )

    monkeypatch.setattr(
        MicroMock,
        '__iter__',
        mock_iter
    )


@pytest.fixture
def mock_repos_iter(monkeypatch):
    """MicroMock.__iter__() mocked to return fake list of repositories."""
    def mock_get_repos(*args, **kwargs):
        return iter(
            [MicroMock(name='repo1'), MicroMock(name='repo2')]
        )

    monkeypatch.setattr(
        MicroMock,
        '__iter__',
        mock_get_repos
    )


@pytest.fixture
def mock_repository(monkeypatch):
    """MockOrganization.get_repo() mocked to raise
    github.GithubException.UnknownObjectException.
    """
    def mock_get_repo(*args, **kwargs):
        raise UnknownObjectException(
            status=401,
            data='Not found',
            headers={}
        )

    monkeypatch.setattr(
        MockOrganization,
        'get_repo',
        mock_get_repo
    )


@pytest.fixture
def client():
    """Return a Client instance."""
    return Client('secret')


@pytest.fixture
def organization():
    """Return an Organization instance."""
    return Organization('Acme')



# Copyright (C) 2020, 2021, 2022, 2023 Serghei Iakovlev <egrep@protonmail.ch>
#
# This file is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This file is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this file.  If not, see <https://www.gnu.org/licenses/>.

"""The main entry point for Gstore.

Invoke as ``gstore`` or ``python -m gstore``.

Functions:

    init() -> None

"""

import sys

from gstore.cli import main


def init() -> None:
    """Run gstore.cli.main() when current file is executed by an interpreter.

    If the file is used as a module, the gstore.cli.main() function will not
    automatically execute. The sys.exit() function is called with a return
    value of gstore.cli.main(), as all good UNIX programs do.
    """
    if __name__ == '__main__':
        sys.exit(main())


init()



# Copyright (C) 2020, 2021, 2022, 2023 Serghei Iakovlev <egrep@protonmail.ch>
#
# This file is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This file is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this file.  If not, see <https://www.gnu.org/licenses/>.

"""
A simple wrapper for github.GitHub to provide convenient way to clone and
sync repositories from the GitHub.
"""

import logging

from github import Github
from github.GithubException import BadCredentialsException
from github.GithubException import UnknownObjectException

from gstore import __version__
from .exceptions import Error
from .models import Organization, Repository

USER_AGENT = f'Gstore/{__version__}'

DEFAULT_HOST = 'api.github.com'
DEFAULT_TIMEOUT = 15


class ValidationError(Error):
    """Base validation error."""


class InvalidCredentialsError(Error):
    """An error resulting from the use of incorrect credentials."""

    def __init__(self):
        message = 'The token provided to access the GitHub API is invalid'
        super().__init__(message)


class Client:
    """A wrapper class around :class:`github.Github` to interact with
    GitHub API.

    :param str token: Authentication token for github.com API requests
    :param str api_host: Default base URL for github.com API requests
    :param int timeout: Timeout for HTTP requests
    :raises ValidationError: in case of GitHub token is not provided.
    """

    def __init__(
            self,
            token: str,
            api_host=DEFAULT_HOST,
            timeout=DEFAULT_TIMEOUT,
    ):
        self.logger = logging.getLogger(f'{__name__}')

        # The use case with Client(token='...', api_host=None)
        if not api_host:
            api_host = DEFAULT_HOST

        # The use case with Client(token=None)
        if not token:
            raise ValidationError(
                'Missing parameter: GitHub Token is required')

        api_url = f'https://{api_host}'
        self.logger.debug('Setting API URL to %s', api_url)

        self.github = Github(
            login_or_token=token,
            base_url=api_url,
            timeout=timeout,
            user_agent=USER_AGENT
        )

    def get_repos(self, org: Organization):
        """Get organization repositories.

        :param Organization org: User's organization
        :return: A collection with repositories
        :rtype: list of :class:`gstore.models.Repository`
        """
        self.logger.info('Getting repositories for organization')

        github_org = self.github.get_organization(org.login)
        repos = github_org.get_repos(
            type='all',
            sort='full_name'
        )

        self.logger.info(
            'Number of available repositories for %s organization: %s',
            org.login,
            repos.totalCount
        )

        retval = []
        for repo in repos:
            retval.append(Repository(repo.name, org))

        return retval

    def resolve_repos(self, repos: list, org: Organization):
        """Resolve repositories from provided list.

        :param list repos: A list of repositories in form 'org:repo'
        :param Organization org: User's organization
        :return: A collection with repositories
        :rtype: list of :class:`gstore.models.Repository`
        """
        self.logger.info('Resolve repositories from provided configuration')

        retval = []

        for name in repos:
            parts = name.split(':')

            if len(parts) != 2 or parts[0] == '' or parts[1] == '':
                self.logger.error(
                    'Invalid repo pattern: "%s", skip resolving',
                    name
                )
                continue

            if parts[0].lower() != org.login.lower():
                continue

            repo_name = parts[-1]

            try:
                # This will do API request, so we'll validate the repo.
                github_org = self.github.get_organization(org.login)
                repo = github_org.get_repo(repo_name)

                retval.append(Repository(repo.name, org))
            except UnknownObjectException:
                self.logger.error('Invalid repository name "%s"', repo_name)
                continue
            except BadCredentialsException as bad_credentials:
                raise InvalidCredentialsError() from bad_credentials

        return retval

    def get_orgs(self):
        """Get organizations for a user.

        :raises InvalidCredentialsError: in case of bad credentials.
        :returns: A collection with organizations
        :rtype: list of :class:`gstore.models.Organization`
        """
        self.logger.info('Getting organizations for user')

        try:
            user = self.github.get_user()
            orgs = user.get_orgs()

            self.logger.info(
                'Number of available organizations for %s user: %s',
                user.login,
                orgs.totalCount
            )

            retval = []
            for org in orgs:
                retval.append(Organization(org.login))

            return retval
        except BadCredentialsException as bad_credentials:
            raise InvalidCredentialsError() from bad_credentials

    def resolve_orgs(self, orgs: list):
        """Resolve organizations from provided list.

        :param list orgs: A list of organizations names
        :raises InvalidCredentialsError: in case of bad credentials.
        :return: A collection with organizations
        :rtype: list of :class:`gstore.models.Organization`
        """
        self.logger.info(
            'Resolve organizations from provided configuration')

        retval = []

        for name in orgs:
            try:
                # This will do API request, so we'll validate the org.
                org = self.github.get_organization(name)
                retval.append(Organization(org.login))
            except UnknownObjectException:
                self.logger.error('Invalid organization name "%s"', name)
                continue
            except BadCredentialsException as bad_credentials:
                raise InvalidCredentialsError() from bad_credentials

        return retval



# Copyright (C) 2020, 2021, 2022, 2023 Serghei Iakovlev <egrep@protonmail.ch>
#
# This file is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This file is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this file.  If not, see <https://www.gnu.org/licenses/>.

"""The CLI entry point. Invoke as `gstore' or `python -m gstore'."""

import logging
import os
import signal
import sys

from .args import argparse
from .client import Client
from .exceptions import Error
from .logger import setup_logger
from .repo import sync


def main():
    """The main function to call gstore from the command line.

    :return: An exit code
    :rtype: int
    """
    args = argparse()
    retval = int(args is None)

    if args:
        setup_logger(verbose=args.verbose, quiet=args.quiet)
        logger = logging.getLogger(__name__)

        try:
            client = Client(token=args.token, api_host=args.host)

            if args.org is None:
                orgs = client.get_orgs()
            else:
                orgs = client.resolve_orgs(args.org)

            base_path = os.path.expanduser(args.target).rstrip('/\\')

            for org in orgs:
                if args.repo is None:
                    repos = client.get_repos(org)
                else:
                    repos = client.resolve_repos(args.repo, org)

                sync(
                    org=org,
                    repos=repos,
                    base_path=base_path,
                    verbose=args.verbose,
                    quiet=args.quiet,
                    jobs=args.jobs,
                )
        except KeyboardInterrupt:  # the user hit control-C
            sys.stderr.write('Received keyboard interrupt, terminating.\n')
            sys.stderr.flush()
            # Control-C is fatal error signal 2, for more see
            # https://tldp.org/LDP/abs/html/exitcodes.html
            retval = 128 + signal.SIGINT
        except Error as gstore_error:
            logger.error(gstore_error)
            retval = 1

    return retval



# Copyright (C) 2020, 2021, 2022, 2023 Serghei Iakovlev <egrep@protonmail.ch>
#
# This file is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This file is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this file.  If not, see <https://www.gnu.org/licenses/>.

"""Repository classes used to wrap git classes for gstore."""

import logging
import multiprocessing
import os
import shutil
from dataclasses import dataclass

import git

from .exceptions import parse_git_errors
from .logger import setup_logger
from .models import Organization, Repository


@dataclass
class _Context:
    """Class for passing a context to parallel processes."""

    base_path: str = ''
    logger: logging.Logger = None


# pylint: disable=invalid-name
_ctx = _Context(logger=logging.getLogger(__name__))


def clone(repo: Repository, target: str):
    """Clone a repository to the target directory."""
    _ctx.logger.info(
        'Clone repository to %s/%s',
        repo.org.login,
        repo.name)

    if os.path.exists(target):
        os.removedirs(target)

    git_url = f'git@github.com:{repo.org.login}/{repo.name}.git'

    try:
        git.Repo.clone_from(
            git_url,
            target,
        )
    except git.GitCommandError as exception:
        _ctx.logger.error('Failed to clone %s/%s', repo.org.login, repo.name)
        for msg in parse_git_errors(exception):
            _ctx.logger.error(msg)


def fetch(repo: Repository, target: str):
    """Sync a repository in the target directory."""
    _ctx.logger.info('Update %s/%s repository', repo.org.login, repo.name)
    local_repo = git.Repo(target)

    if not local_repo.heads:
        _ctx.logger.info(
            'There are no remote branches for %s/%s, skip updating',
            repo.org.login,
            repo.name)
        return

    try:
        _ctx.logger.debug(
            'Download objects and refs from %s/%s repository',
            repo.org.login,
            repo.name)
        local_repo.git.fetch(['--prune', '--quiet'])

        _ctx.logger.debug(
            'Fetch from and integrate with %s/%s repository',
            repo.org.login,
            repo.name)
        local_repo.git.pull(['--all', '--quiet'])
    except git.GitCommandError as exception:
        _ctx.logger.error('Failed to update %s/%s', repo.org.login, repo.name)
        for msg in parse_git_errors(exception):
            _ctx.logger.error(msg)


def _do_sync(repos_list):
    """Perform repos synchronisation. Intended for internal usage."""
    assert isinstance(_ctx.base_path, str) and _ctx.base_path

    for repo in repos_list:
        org_path = os.path.join(_ctx.base_path, repo.org.login)
        repo_path = os.path.join(org_path, repo.name)
        git_path = os.path.join(repo_path, '.git')

        if os.path.exists(repo_path):
            if os.path.isfile(repo_path):
                _ctx.logger.error(
                    'Unable to sync %s. The path %s is a regular file',
                    repo.name,
                    repo_path)
                continue

            if not os.access(repo_path, os.W_OK | os.X_OK):
                _ctx.logger.error(
                    'Unable to sync %s. The path %s is not writeable',
                    repo.name,
                    repo_path)
                continue

            # We're going to run a Git command, but weren't inside a
            # local Git repository.
            if not os.path.exists(git_path):
                _ctx.logger.debug('Remove wrong formed local repo from %s',
                                  repo_path)
                shutil.rmtree(repo_path, ignore_errors=True)

        if os.path.exists(git_path):
            fetch(repo, repo_path)
        else:
            clone(repo, repo_path)


def _init_process(verbose=False, quiet=False, base_path=None):
    """Call when new processes start.

    This function is used as a initializer on a per-process basis due
    to 'spawn' process strategy (at least on Windows and macOs).
    """
    assert isinstance(base_path, str) and base_path

    # Setup logger for use within multiprocessing pool
    setup_logger(verbose, quiet)

    # Setup git base path for use within multiprocessing pool
    _ctx.base_path = base_path

    _ctx.logger = logging.getLogger(f'{__name__}')
    _ctx.logger.info('Initializing process')


def sync(org: Organization, repos: list, base_path: str, **kwargs):
    """
    Sync repositories for an organization.

    :param Organization org: Organization to sync
    :param list repos: Repository list to sync
    :param string base_path: Base target to sync repos
    :keyword bool verbose: Enable debug logging
    :keyword bool quiet: Disable info logging
    :keyword int jobs: the number of worker processes to use
    """
    _ctx.logger.info('Sync repos for %s', org.login)

    org_path = os.path.join(base_path, org.login)

    # Just in case create directories recursively
    if not os.path.exists(org_path):
        _ctx.logger.debug('Creating directory %s', org_path)
        os.makedirs(org_path)

    jobs = kwargs.get('jobs')
    if not jobs:
        jobs = multiprocessing.cpu_count()

    def chunks(lst, n):
        """Yield successive n-sized chunks from lst."""
        for i in range(0, len(lst), n):
            yield lst[i:i + n]

    _ctx.logger.info('Processes to be spawned: %s', jobs)
    with multiprocessing.Pool(processes=jobs, initializer=_init_process,
                              initargs=(kwargs.get('verbose', False),
                                        kwargs.get('quiet', False),
                                        base_path)) as pool:
        pool.map(_do_sync, chunks(repos, jobs))



# Copyright (C) 2020, 2021, 2022, 2023 Serghei Iakovlev <egrep@protonmail.ch>
#
# This file is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This file is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this file.  If not, see <https://www.gnu.org/licenses/>.

from unittest import mock


def test_init():
    """This rather bulky test assures that our init function does everything
    it should:

        - Really calls main if we have '__name__' equals to '__main__'
        - Returns it return value as exit code
        - Does not call main() otherwise

    The final line of the code, the init() call will run at the module import
    time and, therefore, is run at test time.
    """
    from gstore import __main__ as module
    with mock.patch.object(module, 'main', return_value=42):
        with mock.patch.object(module, '__name__', '__main__'):
            with mock.patch.object(module.sys, 'exit') as mock_exit:
                module.init()
                assert mock_exit.call_args[0][0] == 42



# Copyright (C) 2020, 2021, 2022, 2023 Serghei Iakovlev <egrep@protonmail.ch>
#
# This file is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This file is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this file.  If not, see <https://www.gnu.org/licenses/>.

import logging
from unittest import mock

import pytest
from github.GithubException import UnknownObjectException

from gstore.client import Client, InvalidCredentialsError, ValidationError
from gstore.client import DEFAULT_HOST
from gstore.models import Organization, Repository


@pytest.mark.parametrize('token', ['', None, False])
def test_empty_token(token):
    """Call Client() with empty token should raise exception."""
    with pytest.raises(
            ValidationError,
            match='Missing parameter: GitHub Token is required'):
        Client(token)


@pytest.mark.parametrize(
    'provided,url',
    [
        (None, f'https://{DEFAULT_HOST}'),
        ('', f'https://{DEFAULT_HOST}'),
        ('github.example.com', 'https://github.example.com'),
    ]
)
def test_setting_host(provided, url):
    """Call Client() with empty api_host will use default host."""
    with mock.patch.object(logging.Logger, 'debug') as mock_logger:
        Client(token='secret', api_host=provided)
        mock_logger.assert_called_once_with('Setting API URL to %s', url)


def test_resolve_orgs_empty_list(client):
    """Call Client.resolve_orgs() with empty list should return empty list."""
    with mock.patch.object(client.logger, 'info') as mock_logger:
        orgs = client.resolve_orgs([])

        assert len(orgs) == 0
        mock_logger.assert_called_once_with(
            'Resolve organizations from provided configuration')


def test_resolve_orgs_unknown_org(monkeypatch, client):
    """Call Client.resolve_orgs() for a non-existent organization will not stop
    work, but we'll see an error in the logs.
    """
    def mock_get_organization(_):
        # status, data, headers
        raise UnknownObjectException(
            status=401,
            data='Not found',
            headers={}
        )

    monkeypatch.setattr(
        client.github,
        'get_organization',
        mock_get_organization
    )

    with mock.patch.object(client.logger, 'error') as mock_logger:
        orgs = client.resolve_orgs(['foo-bar-baz'])

        assert len(orgs) == 0
        mock_logger.assert_called_once_with(
            'Invalid organization name "%s"', 'foo-bar-baz')


def test_resolve_orgs(mock_micro_organization, client):
    """Call Client.resolve_orgs() with a non-empty list of organization names
    will return a list of :class:`gstore.models.Organization`
    """
    orgs = client.resolve_orgs(['awesome', 'company'])

    assert len(orgs) == 2
    assert isinstance(orgs[0], Organization)
    assert isinstance(orgs[1], Organization)
    assert orgs[0].login == 'awesome'
    assert orgs[1].login == 'company'


def test_resolve_orgs_invalid_token(client):
    """Call Client.resolve_orgs() with invalid token."""
    with pytest.raises(
            InvalidCredentialsError,
            match='The token provided to access the GitHub API is invalid'):
        client.resolve_orgs(['github'])


def test_resolve_repos_empty_list(client, organization):
    """Call Client.resolve_repos() with empty repo list or not expected org
    should return empty list.
    """
    with mock.patch.object(client.logger, 'info') as mock_logger:
        repos = client.resolve_repos([], organization)
        assert len(repos) == 0

        repos = client.resolve_repos(['foo:bar'], organization)
        assert len(repos) == 0

        mock_logger.assert_called_with(
            'Resolve repositories from provided configuration'
        )


def test_get_orgs_empty_list(mock_user, client):
    """Call Client.get_orgs() for a user without organizations will return an
    empty list.
    """
    with mock.patch.object(client.logger, 'info') as mock_logger:
        orgs = client.get_orgs()
        assert len(orgs) == 0

        calls = [
            mock.call('Getting organizations for user'),
            mock.call(
                'Number of available organizations for %s user: %s',
                'mock_user',
                0
            ),
        ]

        mock_logger.assert_has_calls(calls)


def test_get_orgs(mock_user, mock_orgs_iter, client):
    """Call Client.get_orgs() for a user will return a list of
    :class:`gstore.models.Organization`
    """
    orgs = client.get_orgs()

    assert len(orgs) == 2
    assert isinstance(orgs[0], Organization)
    assert isinstance(orgs[1], Organization)
    assert orgs[0].login == 'awesome'
    assert orgs[1].login == 'company'


def test_resolve_repos_unknown_org(mock_organization, mock_repository, client):
    """Call Client.resolve_repos() for a non-existent repository will not stop
    work, but we'll see an error in the logs.
    """
    with mock.patch.object(client.logger, 'error') as mock_logger:
        repos = client.resolve_repos(
            ['Acme:secret-repo'],
            Organization('Acme')
        )

        assert len(repos) == 0
        mock_logger.assert_called_once_with(
            'Invalid repository name "%s"', 'secret-repo')


def test_resolve_repos(mock_organization, client, organization):
    """Call Client.resolve_orgs() with a non-empty list of repo patterns
    will return a list of :class:`gstore.models.Repository`
    """
    repos = client.resolve_repos(['Acme:awesome', 'Acme:repo'], organization)

    assert len(repos) == 2
    assert isinstance(repos[0], Repository)
    assert isinstance(repos[1], Repository)
    assert repos[0].name == 'awesome'
    assert repos[1].name == 'repo'


def test_resolve_repos_invalid_token(client):
    """Call Client.resolve_repos() with invalid token."""
    with pytest.raises(
            InvalidCredentialsError,
            match='The token provided to access the GitHub API is invalid'):
        client.resolve_repos(['Acme:github'], Organization('Acme'))


@pytest.mark.parametrize('repo', ['', ':', 'a', 'a:', 'a:b:', 'a:b:c'])
def test_resolve_repos_invalid_repo_pattern(repo, client, organization):
    """Call Client.resolve_repos() with invalid repo pattern
    should return empty list and log error.
    """
    with mock.patch.object(client.logger, 'error') as mock_logger:
        repos = client.resolve_repos([repo], organization)

        assert len(repos) == 0
        mock_logger.assert_called_once_with(
            'Invalid repo pattern: "%s", skip resolving', repo)


def test_get_repos_empty_list(mock_organization, client, organization):
    """Call Client.get_repos() for an organization without repositories will
    return an empty list.
    """
    with mock.patch.object(client.logger, 'info') as mock_logger:
        repos = client.get_repos(organization)
        assert len(repos) == 0

        calls = [
            mock.call('Getting repositories for organization'),
            mock.call(
                'Number of available repositories for %s organization: %s',
                'Acme',
                0
            ),
        ]

        mock_logger.assert_has_calls(calls)


def test_get_repos(mock_organization, mock_repos_iter, client):
    """Call Client.get_repos() for an organization will return a list of
    gstore.models.Repository.
    """
    repos = client.get_repos(Organization('awesome'))

    assert len(repos) == 2
    assert isinstance(repos[0], Repository)
    assert isinstance(repos[1], Repository)
    assert repos[0].name == 'repo1'
    assert repos[1].name == 'repo2'



# Copyright (C) 2020, 2021, 2022, 2023 Serghei Iakovlev <egrep@protonmail.ch>
#
# This file is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This file is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this file.  If not, see <https://www.gnu.org/licenses/>.

import logging
import signal
from unittest import mock

from gstore import cli, client


def test_main_keyboard_interrupt(monkeypatch):
    """Handle Ctrl-C keyboard event and gracefully terminate work."""
    def mock_init(*args, **kwargs):
        raise KeyboardInterrupt

    monkeypatch.setattr('sys.argv', ['gstore', '--token', 'secret'])
    monkeypatch.setattr(client.Client, '__init__', mock_init)

    assert cli.main() == 128 + signal.SIGINT


def test_get_orgs_without_token(monkeypatch):
    """Call with invalid token leads to exit with a status of one."""
    monkeypatch.setattr('sys.argv', ['gstore', '--token', 'secret'])

    with mock.patch.object(logging.Logger, 'info') as mock_logger:
        assert cli.main() == 1

        mock_logger.assert_called_once_with(
            'Getting organizations for user')


def test_resolve_orgs_without_token(monkeypatch):
    """Call with invalid token leads to exit with a status of one."""
    monkeypatch.setattr('sys.argv', [
        'gstore',
        '--token', 'secret',
        '--org', 'Acme'
    ])

    with mock.patch.object(logging.Logger, 'info') as mock_logger:
        assert cli.main() == 1

        mock_logger.assert_called_once_with(
            'Resolve organizations from provided configuration')



# Copyright (C) 2020, 2021, 2022, 2023 Serghei Iakovlev <egrep@protonmail.ch>
#
# This file is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This file is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this file.  If not, see <https://www.gnu.org/licenses/>.

import pytest

from gstore import env


@pytest.mark.parametrize(
    'provided,expected',
    [
        ([('GH_TOKEN', 'secret1')], 'secret1'),
        ([('GITHUB_TOKEN', 'secret2')], 'secret2'),
        ([('GITHUB_TOKEN', 'secret3'),
          ('GH_TOKEN', 'secret4')], 'secret4'),
        ([('GH_TOKEN', 'secret5'),
          ('GITHUB_TOKEN', 'secret6')], 'secret5'),
        ([('GH_ENTERPRISE_TOKEN', 'secret7')], 'secret7'),
        ([('GITHUB_ENTERPRISE_TOKEN', 'secret8')], 'secret8'),
        ([('GITHUB_ENTERPRISE_TOKEN', 'secret9'),
          ('GH_ENTERPRISE_TOKEN', 'secret10')], 'secret10'),
        ([('GH_ENTERPRISE_TOKEN', 'secret11'),
          ('GITHUB_ENTERPRISE_TOKEN', 'secret12')], 'secret11'),
        ([('GH_TOKEN', '')], None),
        ([('GITHUB_TOKEN', '')], None),
        ([('GH_ENTERPRISE_TOKEN', '')], None),
        ([('GITHUB_ENTERPRISE_TOKEN', '')], None),
        ([('GH_TOKEN', ''),
          ('GITHUB_TOKEN', ''),
          ('GH_ENTERPRISE_TOKEN', ''),
          ('GITHUB_ENTERPRISE_TOKEN', '')], None),
    ])
def test_lookup_token(provided, expected, monkeypatch):
    """Call gstore.env.lookup_token() will return a token from environment
    variables (if any is set), taking into account the priority of these
    variables.
    """
    for k, v in provided:
        monkeypatch.setenv(k, v)

    assert env.lookup_token() == expected


@pytest.mark.parametrize(
    'provided,expected',
    [
        (None, None),
        ('', None),
        ('secret', 'secret'),
    ]
)
def test_get_host(provided, expected, monkeypatch):
    """Call gstore.env.get_host() will return a host if environment
    variable is set with not empty string, otherwise None.
    """
    if provided is None:
        monkeypatch.delenv('GH_HOST', raising=False)
    else:
        monkeypatch.setenv('GH_HOST', provided)

    assert env.get_host() == expected


@pytest.mark.parametrize(
    'provided,expected',
    [
        (None, None),
        ('', None),
        ('~/backup', '~/backup'),
    ]
)
def test_get_target(provided, expected, monkeypatch):
    """Call gstore.env.get_target() will return a target if environment
    variable is set with not empty string, otherwise None.
    """
    if provided is None:
        monkeypatch.delenv('GSTORE_DIR', raising=False)
    else:
        monkeypatch.setenv('GSTORE_DIR', provided)

    assert env.get_target() == expected



# Copyright (C) 2020, 2021, 2022, 2023 Serghei Iakovlev <egrep@protonmail.ch>
#
# This file is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This file is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this file.  If not, see <https://www.gnu.org/licenses/>.

"""The top-level module for Gstore.

This module tracks the version of the package as well as the base
package info used by various functions within Gstore.

Modules:

    __main__
    args
    cli
    client
    env
    exceptions
    logger
    models
    repo

Misc variables:

    __copyright__
    __version__
    __license__
    __author__
    __author_email__
    __url__
    __description__

Refer to the `documentation <https://gstore.readthedocs.io/>`_ for
details on the use of this package.

"""

__copyright__ = 'Copyright (C) 2020, 2021, 2022, 2023 Serghei Iakovlev'
__version__ = '0.6.1'
__license__ = 'GPLv3+'
__author__ = 'Serghei Iakovlev'
__author_email__ = 'egrep@protonmail.ch'
__url__ = 'https://gstore.readthedocs.io'
__description__ = 'Synchronize GitHub repositories made easy.'



# Copyright (C) 2020, 2021, 2022 Serghei Iakovlev <egrep@protonmail.ch>
#
# This file is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This file is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this file.  If not, see <https://www.gnu.org/licenses/>.

"""Setup module for Gstore."""

import codecs
import re
from os import path

from setuptools import find_packages, setup


def read_file(filepath):
    """Read content from a UTF-8 encoded text file."""
    with codecs.open(filepath, 'rb', 'utf-8') as file_handle:
        return file_handle.read()


PKG_NAME = 'gstore'
PKG_DIR = path.abspath(path.dirname(__file__))
META_PATH = path.join(PKG_DIR, PKG_NAME, '__init__.py')
META_CONTENTS = read_file(META_PATH)


def load_long_description():
    """Load long description from file README.rst."""
    def changes():
        changelog = path.join(PKG_DIR, 'CHANGELOG.rst')
        pat = r"(\d+.\d+.\d+ \(.*?\)\r?\n.*?)\r?\n\r?\n\r?\n----\r?\n\r?\n\r?\n"  # noqa: E501
        result = re.search(pat, read_file(changelog), re.S)

        return result.group(1) if result else ''

    try:
        title = f"{PKG_NAME}: {find_meta('description')}"
        head = '=' * (len(title) - 1)

        contents = (
            head,
            format(title.strip(' .')),
            head,
            read_file(path.join(PKG_DIR, 'README.rst')).split(
                '.. teaser-begin'
            )[1],
            '',
            read_file(path.join(PKG_DIR, 'CONTRIBUTING.rst')),
            '',
            'Release Information',
            '===================\n',
            changes(),
            '',
            f"`Full changelog <{find_meta('url')}/en/latest/changelog.html>`_.",  # noqa: E501
            '',
            read_file(path.join(PKG_DIR, 'SECURITY.rst')),
            '',
            read_file(path.join(PKG_DIR, 'AUTHORS.rst')),
        )

        return '\n'.join(contents)
    except (RuntimeError, FileNotFoundError) as read_error:
        message = 'Long description could not be read from README.rst'
        raise RuntimeError(f'{message}: {read_error}') from read_error


def is_canonical_version(version):
    """Check if a version string is in the canonical format of PEP 440."""
    pattern = (
        r'^([1-9][0-9]*!)?(0|[1-9][0-9]*)(\.(0|[1-9][0-9]*))'
        r'*((a|b|rc)(0|[1-9][0-9]*))?(\.post(0|[1-9][0-9]*))'
        r'?(\.dev(0|[1-9][0-9]*))?$')
    return re.match(pattern, version) is not None


def find_meta(meta):
    """Extract __*meta*__ from META_CONTENTS."""
    meta_match = re.search(
        r"^__{meta}__\s+=\s+['\"]([^'\"]*)['\"]".format(meta=meta),
        META_CONTENTS,
        re.M
    )

    if meta_match:
        return meta_match.group(1)
    raise RuntimeError(
        f'Unable to find __{meta}__ string in package meta file')


def get_version_string():
    """Return package version as listed in `__version__` in meta file."""
    # Parse version string
    version_string = find_meta('version')

    # Check validity
    if not is_canonical_version(version_string):
        message = (
            'The detected version string "{}" is not in canonical '
            'format as defined in PEP 440.'.format(version_string))
        raise ValueError(message)

    return version_string


# What does this project relate to.
KEYWORDS = [
    'git',
    'github',
    'backup',
    'repo',
    'sync',
]

# Classifiers: available ones listed at https://pypi.org/classifiers
CLASSIFIERS = [
    'Development Status :: 5 - Production/Stable',

    'Environment :: Console',

    'Intended Audience :: Developers',
    'Intended Audience :: System Administrators',
    'Intended Audience :: Information Technology',

    'Natural Language :: English',

    'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',  # noqa: E501
    'Operating System :: OS Independent',

    'Programming Language :: Python',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.8',
    'Programming Language :: Python :: 3.9',
    'Programming Language :: Python :: 3.10',
    'Programming Language :: Python :: 3.11',
    'Programming Language :: Python :: 3 :: Only',

    'Topic :: System :: Archiving :: Backup',
    'Topic :: System :: Software Distribution',
    'Topic :: Software Development :: Build Tools',
    'Topic :: Software Development :: Version Control',
    'Topic :: Software Development :: Version Control :: Git',
]

# Dependencies that are downloaded by pip on installation and why.
INSTALL_REQUIRES = [
    'pygithub>=1.55',  # Interact with GitHub objects
    'gitpython>=3.0.6',  # Interact with Git repositories
]

DEPENDENCY_LINKS = []

# List additional groups of dependencies here (e.g. testing dependencies).
# You can install these using the following syntax, for example:
#
#    $ pip install -e .[testing,docs,develop]
#
EXTRAS_REQUIRE = {
    # Dependencies that are required to run tests
    'testing': [
        'coverage[toml]>=6.0',  # Code coverage measurement for Python
        'pytest>=6.2.4',  # Our tests framework
        'pylint>=2.6.0,!=2.6.1',  # Python code static checker
        'flake8>=6.0.0',  # The modular source code checker
        'flake8-import-order>=0.18.1',  # Checks the ordering of imports
        'flake8-blind-except>=0.2.0',  # Checks for blind except: statements
        'check-manifest>=0.45',  # Check MANIFEST.in
    ],
    'docs': [
        'sphinx>=3.5.0',  # Python documentation generator,
        'furo>=2022.9.29',  # Sphinx documentation theme
    ],
}

# Dependencies that are required to develop package
DEVELOP_REQUIRE = [
    'twine>=3.3.0',  # Publishing packages on PyPI
    'setuptools>=53.0.0',  # Build and install packages
    'wheel>=0.36.2',  # A built-package format for Python
    'check-wheel-contents>=0.2.0',  # Check wheels have the right contents
]

EXTRAS_REQUIRE['develop'] = \
    DEVELOP_REQUIRE + \
    EXTRAS_REQUIRE['testing'] + \
    EXTRAS_REQUIRE['docs']

# Project's URLs
PROJECT_URLS = {
    'Documentation': 'https://gstore.readthedocs.io',
    'Changelog': f"{find_meta('url')}/en/latest/changelog.html",
    'Bug Tracker': 'https://github.com/sergeyklay/gstore/issues',
    'Source Code': 'https://github.com/sergeyklay/gstore',
}

ENTRY_POINTS = {
    'console_scripts': [
        f'{PKG_NAME}={PKG_NAME}.cli:main'
    ]
}

if __name__ == '__main__':
    setup(
        name=PKG_NAME,
        version=get_version_string(),
        author=find_meta('author'),
        author_email=find_meta('author_email'),
        maintainer=find_meta('author'),
        maintainer_email=find_meta('author_email'),
        license=find_meta('license'),
        description=find_meta('description'),
        long_description=load_long_description(),
        long_description_content_type='text/x-rst',
        keywords=KEYWORDS,
        url=find_meta('url'),
        project_urls=PROJECT_URLS,
        classifiers=CLASSIFIERS,
        packages=find_packages(exclude=['tests.*', 'tests']),
        platforms='any',
        include_package_data=True,
        zip_safe=False,
        python_requires='>=3.8, <4',
        install_requires=INSTALL_REQUIRES,
        dependency_links=DEPENDENCY_LINKS,
        extras_require=EXTRAS_REQUIRE,
        entry_points=ENTRY_POINTS,
    )



# Copyright (C) 2020, 2021, 2022, 2023 Serghei Iakovlev <egrep@protonmail.ch>
#
# This file is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This file is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this file.  If not, see <https://www.gnu.org/licenses/>.

"""Helper routines to work with environment variables.

Provides a convenient way to work with environment variables
used by various functions within Gstore.

Functions:

    lookup_token() -> str or None
    get_host() -> str or None
    get_target() -> str or None

Data:

    TOKEN_NAMES

"""

import os

TOKEN_NAMES = (
    'GH_TOKEN',
    'GITHUB_TOKEN',
    'GH_ENTERPRISE_TOKEN',
    'GITHUB_ENTERPRISE_TOKEN',
)


def lookup_token() -> str or None:
    """Lookup a personal access token from environment variables.

    This function looks for token variable in the following order:

        1. GH_TOKEN
        2. GITHUB_TOKEN
        3. GH_ENTERPRISE_TOKEN
        4. GITHUB_ENTERPRISE_TOKEN

    :returns: A personal access token if any or None
    :rtype: str or None
    """
    for name in TOKEN_NAMES:
        token = os.environ.get(name)
        if token:
            return token

    return None


def get_host() -> str or None:
    """Get GitHub API hostname from GH_HOST environment variable.

    This function may return None if there is no environment
    variable, or it is empty.

    :returns: The GitHub API hostname if any or None
    :rtype: str or None
    """
    return os.environ.get('GH_HOST') or None


def get_target() -> str or None:
    """Get base target to sync repos.

    This function may return None if there is no environment
    variable, or it is empty.

    :returns: The path to base directory if any or None
    :rtype: str or None
    """
    return os.environ.get('GSTORE_DIR') or None



# Copyright (C) 2020, 2021, 2022, 2023 Serghei Iakovlev <egrep@protonmail.ch>
#
# This file is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This file is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this file.  If not, see <https://www.gnu.org/licenses/>.

"""Helper routines and classes to work exceptions.

Classes:

    Error

Functions:

    parse_git_errors() -> list

"""

import re

import git


def parse_git_errors(ex: git.CommandError) -> list:
    """Parse errors produced by :class:`git.GitCommandError` subclasses.

    :param git.GitCommandError ex: An error instance
    :return: A list of error messages
    :rtype: list of str
    """
    def replace(msg):
        """Replace multiline messages to a single-line form."""
        oneliner = re.compile(r'[\s]{2,}|\n')
        cleaner = re.compile(r"(?:(?:std(?:out|err)|error):[ ]*'?)")

        return cleaner.sub('', oneliner.sub(' ', msg)).strip(" .'")

    holders = []

    if ex.stdout:
        holders.append(ex.stdout)
    if ex.stderr:
        holders.append(ex.stderr)

    messages = map(replace, holders)

    return list(messages)


class Error(Exception):
    """Base class for exceptions in gstore module."""
